import unittest

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import missings, msa


class ConstSymbolProviderTests(unittest.TestCase):

    def test_1_no_symbol_provided(self):
        missing_symbol = missings.MissingBase()
        const_symbol_provider = missings.ConstBaseProvider(missing_symbol)

        expected_symbol = graph.Base('?')
        actual_symbol = const_symbol_provider.get_base(msa.SequenceID('s'), 0)
        self.assertEqual(expected_symbol, actual_symbol)

    def test_2_symbol_provided(self):
        const_symbol_provider = missings.ConstBaseProvider(missings.MissingBase('*'))

        expected_symbol = graph.Base('*')
        actual_symbol = const_symbol_provider.get_base(msa.SequenceID('s'), 0)
        self.assertEqual(expected_symbol, actual_symbol)

    def test_3_incorrect_missing_symbol(self):
        with self.assertRaises(ValueError) as e:
            _ = missings.MissingBase('**')

        expected_message = 'Missing symbol must be a single character.'
        actual_message = str(e.exception)
        self.assertEqual(expected_message, actual_message)


if __name__ == '__main__':
    unittest.main()



"""AffinityTree and related objects definitions."""

import math
from typing import List, NewType, Dict, Optional

import newick
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa

AffinityNodeID = NewType('AffinityNodeID', int)


class AffinityNode(object):
    """Node of Affinity Tree.

    Args:
        id_: Affinity Node ID.
        parent: ID of the parent node.
        children: IDs of the children nodes.
        sequences: IDs of the _sequences assigned to this node.
        mincomp: Minimum from the compatibilities of this consensus to the
            assigned sequences
        compatibilities: Dictionary of compatibilities to all sequences.
        consensus: Path of the consensus defined as path in Poagraph.

    Attributes:
        id_ (AffinityNodeID): ID of this node
        parent (AffinityNodeID): ID of the parent node
        children(List[AffinityNodeID]): IDs of the children nodes.
        sequences(List[SequenceID]): IDs of the sequences assigned to the node.
        mincomp(Compatibility): Minimum from the compatibilities of this
            consensus to the assigned sequences
        compatibilities(Dict[SequenceID, Compatibility]): Dictionary of
            compatibilities to any _sequences.
        consensus(SeqPath): Path of the consensus defined as path in Poagraph.
    """

    def __init__(self,
                 id_: AffinityNodeID,
                 parent: Optional[AffinityNodeID] = None,
                 children: Optional[List[AffinityNodeID]] = None,
                 sequences: Optional[List[msa.SequenceID]] = None,
                 mincomp: Optional[graph.Compatibility] = None,
                 compatibilities: Optional[Dict[msa.SequenceID,
                                                graph.Compatibility]] = None,
                 consensus: Optional[graph.SeqPath] = None):
        self.id_: AffinityNodeID = id_
        self.parent: AffinityNodeID = parent
        self.children: List[AffinityNodeID] = children if children else []
        self.sequences: List[msa.SequenceID] = sequences if sequences else []
        self.mincomp: graph.Compatibility = mincomp if mincomp else graph.Compatibility(0)
        self.compatibilities: Dict[msa.SequenceID,
                                   graph.Compatibility] = compatibilities if compatibilities else {}
        self.consensus: graph.SeqPath = consensus

    def __str__(self):
        return f"ID: {self.id_}, "\
               f"parent: {self.parent}, " \
               f"children: {self.children}, " \
               f"mincomp: {str(self.mincomp)}, " \
               f"consensus length: {len(self.consensus)}, "\
               f"_sequences: {self.sequences}."

    def __eq__(self, other):
        return self.parent == other.parent and \
               self.id_ == other.id and \
               self.sequences == other.sequences and \
               math.isclose(self.mincomp.value, other.mincomp.value) and \
               self.compatibilities == other.compatibilities and \
               self.consensus == other.consensus


class AffinityTree(object):
    """
    Affinity Tree defined as list of nodes.

    No correctness checking is performed.

    Args:
        nodes: Nodes of the tree.

    Attributes:
        nodes (List[AffinityNode]): All nodes of the tree. Relations between
            nodes are described by their attributes.
    """

    def __init__(self, nodes: Optional[List[AffinityNode]] = None):
        self.nodes: List[AffinityNode] = nodes if nodes else []

    def get_node(self, id_: AffinityNodeID) -> AffinityNode:
        """Returns affinity node with given ID.

        Args:
            id_: ID of the node to return.

        Returns:
            Affinity Node of given ID.

        Raises:
            KeyError: No node with given ID exists in this Affinity Tree.
        """

        for node in self.nodes:
            if node.id_ == id_:
                return node
        raise KeyError("No node with given ID.")

    def get_max_node_id(self) -> AffinityNodeID:
        """Returns the largest number used as the tree node ID.

        Returns:
            The largest AffinityNodeID used in this Affinity Tree. Returns
                AffinityNodeID(-1) if the tree has no nodes.
        """

        if len(self.nodes) == 0:
            return AffinityNodeID(-1)
        return max([node.id_ for node in self.nodes])

    def as_newick(self,
                  seq_id_to_metadata: Dict[msa.SequenceID,
                                           graph.SequenceMetadata] = None,
                  separate_leaves=False) -> str:
        """Returns Affinity Tree in Newick format.

        Args:
            seq_id_to_metadata: Dictionary of _sequences IDs to the desired
                name used in newick file. For example:
                                {SequenceID('KM0123'): 'cat',
                                SequenceID('ZX124'): 'dog'}
            separate_leaves: A switch to control if tree leaves having
                assigned multiple _sequences should have appended
                children singleton leaves single sequence assigned.

        Returns:
            A string with the Affinity Tree converted to newick format.
            https://en.wikipedia.org/wiki/Newick_format
            If the tree has no nodes, an empty string is returned.
        """

        def _get_sequence_attr_if_exists(seq_metadata: graph.SequenceMetadata,
                                         attr: str) -> str:
            """Returns dictionary value if they key attr exists."""

            if attr in seq_metadata:
                return str(seq_metadata[attr])
            else:
                return ""

        def _newick_nhx(newick_node: newick.Node) -> str:
            """Converts newick tree to newick string"""

            node_label = newick_node.name or ''
            if newick_node._length:
                for cn in sorted_nodes:
                    if str(cn.id_) == newick_node.name:
                        if seq_id_to_metadata:
                            if len(cn.sequences) == 1:
                                name = _get_sequence_attr_if_exists(seq_id_to_metadata[cn.sequences[0]], "name")
                                if name == "":
                                    name = cn.sequences[0]
                                group = _get_sequence_attr_if_exists(seq_id_to_metadata[cn.sequences[0]], "group")
                                seqid = cn.sequences[0]
                                metadata = f"[&&NHX:name={name}:group={group}:seqid={seqid}:mincomp={cn.mincomp}]"
                            elif len(cn.sequences) == 0:
                                name = f"EmptyAffinityNode {cn.id_}"
                                metadata = f"[&&NHX:name={name}:mincomp={cn.mincomp}]"
                            else:
                                name = f"AffinityNode {cn.id_}"
                                metadata = f"[&&NHX:name={name}:mincomp={cn.mincomp}]"
                        else:
                            if len(cn.sequences) == 1:
                                name = cn.sequences[0]
                            elif len(cn.sequences) == 0:
                                name = f"EmptyAffinityNode {cn.id_}"
                            else:
                                name = f"AffinityNode {cn.id_}"
                            mincomp = cn.mincomp
                            metadata = f"[&&NHX:name={name}:mincomp={mincomp}]"
                try:
                    node_label += ':' + newick_node._length + metadata
                except Exception:
                    print("metadata")
            descendants = ','.join([_newick_nhx(n)
                                    for n in newick_node.descendants])
            if descendants:
                descendants = '(' + descendants + ')'
            return descendants + node_label

        if not self.nodes:
            return ""

        sorted_nodes = sorted(self.nodes, key=lambda x: x.id_)

        if separate_leaves:
            new_leaves_count = 0
            for node in self.nodes:
                if len(node.children) == 0 and len(node.sequences) > 1:
                    for seq_id in node.sequences:
                        affinity_node_id = len(self.nodes) + new_leaves_count
                        node.children.append(affinity_node_id)
                        sorted_nodes.append(AffinityNode(id_=AffinityNodeID(affinity_node_id),
                                                         parent=node.id_,
                                                         children=[],
                                                         sequences=[seq_id],
                                                         mincomp=graph.Compatibility(1.0)
                                                         ))
                        new_leaves_count += 1

        nodes_to_process = [(None, sorted_nodes[0])]
        newick_tree = None
        while nodes_to_process:
            n = nodes_to_process.pop()
            node_parent_label = n[0]
            node = n[1]

            label = str(node.id_)
            if node.parent is None:
                length = "1"
            else:
                parent_minComp = sorted_nodes[node.parent].mincomp.base_value().value
                length = str((1 - parent_minComp) - (1 - node.mincomp.base_value().value))

            newick_node = newick.Node(name=label, length=length)

            if newick_tree is None:
                newick_tree = newick_node
            else:
                parent_node = newick_tree.get_node(node_parent_label)
                parent_node.add_descendant(newick_node)

            for child in node.children:
                nodes_to_process.append((label, sorted_nodes[child]))
        return "(" + _newick_nhx(newick_tree) + ")"



from typing import List, Tuple, Callable

from pangtreebuild.pangenome import graph


class NodePO:
    """Poagraph node representation in po file."""

    def __init__(self,
                 base: bytes,
                 aligned_to: graph.NodeID,
                 in_nodes: List[graph.NodeID],
                 sequences_ids: List[int]):
        self.base = base
        self.aligned_to = aligned_to
        self.in_nodes = in_nodes
        self.sequences_ids = sequences_ids


class SequencePO:
    """Sequences representation in po file."""

    def __init__(self,
                 name: str,
                 nodes_count: int,
                 weight: int,
                 consensus_id: int,
                 start_node_id: graph.NodeID):
        self.name = name
        self.nodes_count = nodes_count
        self.weight = weight
        self.consensus_id = consensus_id
        self.start_node_id = start_node_id


def poagraph_to_PangenomePO(poagraph: graph.Poagraph) -> str:
    """Converts poagraph to .po file."""

    po_nodes, po_sequences = _convert_to_po_input_data(poagraph)
    return poagraph_elements_to_PangenomePO(po_nodes,
                                            po_sequences,
                                            poagraph.datatype)


def poagraph_elements_to_PangenomePO(po_nodes: List[NodePO],
                                     po_sequences: List[SequencePO],
                                     datatype: graph.DataType) -> str:
    """Converts po specific elements to po file."""

    introduction_lines = "\n".join(_get_introduction_lines(len(po_nodes),
                                                           len(po_sequences)))
    sequences_lines = "\n".join(_get_sequences_lines(po_sequences))
    nodes_lines = "\n".join(_get_nodes_lines(po_nodes, datatype))
    return "\n".join([introduction_lines, sequences_lines, nodes_lines])


def _convert_to_po_input_data(p: graph.Poagraph) -> \
        Tuple[List[NodePO], List[SequencePO]]:
    po_nodes = []
    po_sequences = []
    sequences_weights = p.get_sequences_weights(p.get_sequences_ids())

    for node in p.nodes:
        po_nodes.append(NodePO(base=node._base.value,
                               aligned_to=node.aligned_to,
                               in_nodes=set(),
                               sequences_ids=[]))

    seq_int_id = -1
    for seq_id, sequence in p.sequences.items():
        nodes_count = sum([len(path) for path in sequence.paths])
        if nodes_count == 0:
            continue
        seq_int_id += 1
        po_sequences.append(SequencePO(name=str(seq_id),
                                       nodes_count=nodes_count,
                                       weight=sequences_weights[seq_id],
                                       consensus_id=-1,
                                       start_node_id=sequence.paths[0][0]))
        for path in sequence.paths:
            previous_node_id = None
            for node_id in path:
                po_nodes[node_id].sequences_ids.append(seq_int_id)
                if previous_node_id is not None:
                    po_nodes[node_id].in_nodes.add(previous_node_id)
                previous_node_id = node_id

    for node in po_nodes:
        node.in_nodes = list(node.in_nodes)

    return po_nodes, po_sequences


def _get_introduction_lines(nodes_count: int, sequences_count: int) -> \
        List[str]:
    introduction_lines = ["VERSION=pangenome",
                          "NAME=pangenome",
                          "TITLE=pangenome",
                          f"LENGTH={nodes_count}",
                          f"SOURCECOUNT={sequences_count}"]
    return introduction_lines


def _get_sequences_lines(po_sequences: List[SequencePO]) -> List[str]:
    if not po_sequences:
        raise Exception("No _sequences info to write in PO file.")
    i = -2
    po_sequences_lines = [None] * len(po_sequences) * 2
    for sequence in po_sequences:
        i += 2
        po_sequences_lines[i] = f"SOURCENAME={sequence.name}"
        po_sequences_lines[i + 1] = ("SOURCEINFO=" +
                                     " ".join([f"{sequence.nodes_count}",
                                               f"{sequence.start_node_id}",
                                               f"{sequence.weight}",
                                               f"{sequence.consensus_id}",
                                               f"{sequence.name}"]))
    return po_sequences_lines


def _get_nodes_lines(po_nodes: List[NodePO], datatype: graph.DataType) -> \
        List[str]:
    if not po_nodes:
        raise Exception("No nodes info to write in PO file.")
    i = 0

    _get_node_code = _get_node_code_conversion_function(datatype)
    po_nodes_lines = [None] * len(po_nodes)
    for node in po_nodes:
        po_nodes_lines[i] = "".join([_get_node_code(node.base),
                                     ":",
                                     _get_in_nodes_info(node.in_nodes),
                                     _get_sources_info(node.sequences_ids),
                                     _get_aligned_to_info(node.aligned_to)])
        i += 1
    return po_nodes_lines


def _get_in_nodes_info(in_nodes: List[graph.NodeID]) -> str:
    return "".join([f'L{i}' for i in in_nodes])


def _get_sources_info(sources_ids: List[int]) -> str:
    return "".join([f'S{i}' for i in sources_ids])


def _get_aligned_to_info(aligned_to: graph.NodeID) -> str:
    return f"A{aligned_to}" if aligned_to is not None else ""


def _get_node_code_conversion_function(datatype: graph.DataType) -> \
        Callable[[bytes], str]:
    if datatype.value == graph.DataType.Proteins.value:
        return _get_protein_node_code
    elif datatype.value == graph.DataType.Nucleotides.value:
        return _get_nucleotides_node_code
    else:
        raise Exception("Unknown data type. Cannot create PO file.")


def _get_protein_node_code(base: bytes) -> str:
    return base.decode("ASCII").upper()


def _get_nucleotides_node_code(base: bytes) -> str:
    return base.decode("ASCII").lower()



from datetime import datetime
from io import StringIO
import os
from pathlib import Path
import shutil
from typing import Optional


def get_file_content(path: Path) -> str:
    """Returns file content."""

    with open(path) as input_file:
        return input_file.read()


def get_file_content_stringio(path: Path) -> StringIO:
    """Returns file content."""

    return StringIO(get_file_content(path))


def get_child_path(directory: Path, child_name: str) -> Path:
    """Returns path of file child_name placed in directory."""

    return directory.joinpath(child_name)


def dir_exists(dir_path: Path):
    """Returns True if directory exists, False otherwise."""

    return dir_path.exists() and dir_path.is_dir()


def file_exists(file_path: Path):
    """Returns True if file exists, False otherwise."""

    return file_path.is_file()


def create_dir(dir_path: Path):
    """Creates dir as specified by dir_path."""

    if not dir_exists(dir_path):
        dir_path.mkdir()


def get_cwd() -> Path:
    """Returns current working directory."""

    return Path(os.getcwd())


def get_current_time() -> str:
    """Returns current date and time in format MM_DD__HH_MM_SS"""

    return datetime.now().strftime('%m_%d__%H_%M_%S')


def save_to_file(filecontent: str,
                 filename: Path,
                 mode: Optional[str] = 'w') -> \
        None:
    """Saves string to file."""

    with open(filename, mode) as output:
        output.write(filecontent)


def get_child_dir(parent_path: Path, child_dir_name: str):
    """Creates dir if not exists and returns its path."""

    child_dir_path = parent_path.joinpath(child_dir_name)
    if not child_dir_path.is_dir():
        child_dir_path.mkdir()
    return child_dir_path.resolve()


def remove_dir(dir_path: Path) -> None:
    """Removes directory."""

    if dir_path.exists() and dir_path.is_dir():
        shutil.rmtree(dir_path)



# -*- coding: utf-8 -*-
import sys
sys.path.append('..')
from ..mafreader import start_position

class EdgeInfo:
    def __init__(self, seq_id, start_pos):
        self.seq_id = seq_id
        self.start = start_pos
        
    def set_start_position(self, alignment, orientation):
        for seq in alignment:
            if seq.id == self.seq_id and start_position(seq) == self.start:
                if orientation*seq.annotations["strand"] == 1:
                    break
                elif orientation == -1:
                    self.start = seq.annotations["srcSize"] - seq.annotations["start"] - seq.annotations["size"]
                    break
                else:
                    self.start = seq.annotations["start"]
                    break



"""Parameters of the Affinity Tree generation algorithms."""

from io import StringIO
from pathlib import Path
from typing import Union


class Blosum(object):
    """BLOSUM matrix file content and its full path.

    This file is used by the poa software to determine consensuses paths
    in poagraph. The matrix in this file must contain symbol used by
    pangtreebuild for missing nucleotides/proteins. Lower-case is interpreted
    as nucleotides and upper-case as proteins.

    Args:
        file_content: Blosum file content required to validate its content.
        filepath: Path to the Blosum file, required by the poa software.

    Raises:
        ValueError: If file is empty or the BLOSUM matrix cannot be found.

    Attributes:
        filepath (Path): Full path to the blosum file.
            Required by poa software.
    """

    def __init__(self, file_content: StringIO, filepath: Path):
        self.filepath = filepath
        self._blosum_lines = file_content.readlines()
        self._blosum_symbols_line = None
        self._read_and_raise_exception_if_incorrect()

    def _read_and_raise_exception_if_incorrect(self) -> None:
        """Blosum file validation.

        Raises:
            ValueError: If file is empty or the BLOSUM matrix cannot be found.
        """

        if not self._blosum_lines:
            raise ValueError("""Empty blosum file. Provide a valid blosum file
                                or use te default one.""")

        blosum_symbols_line = None
        for blosum_line in self._blosum_lines:
            if len(blosum_line) > 0 and blosum_line[0] == " ":
                blosum_symbols_line = blosum_line
                break
        if not blosum_symbols_line:
            raise ValueError("""Cannot find the horizontal line of symbols
                                in blosum file. It should be the first line
                                beginning with a whitespace.""")
        self._blosum_symbols_line = blosum_symbols_line

    def check_if_symbol_is_present(self, missing_symbol: str) -> bool:
        """Checks if missing_base is present in the matrix.

        Args:
            missing_symbol: The symbol to be searched for in the BLOSUM matrix.

        Returns:
            bool: True if the symbol is included in the BLOSUM matrix.
                Otherwise an exception is raised.

        Raises:
            ValueError: If missing_base has length other than 1 or the symbol
                is not included in the BLOSUM matrix.
        """

        blosum_symbols = str.strip(self._blosum_symbols_line).split(" ")
        if len(missing_symbol) != 1:
            raise ValueError(f"""The missing symbol must be single character.
                                 Got{len(missing_symbol)} instead.""")
        if missing_symbol in blosum_symbols:
            return True
        else:
            raise ValueError("""Cannot find symbol used as missing base
                                in blosum file.""")


class Hbmin(object):
    """(Poa parameter) Minimum value of sequence compatibility to consensus.

    Args:
        value: Hbmin value. Must be in range [0,1].

    Raises:
        ValueError: If hbmin value is not in [0,1].
        TypeError: If value is not floatable.

    Attributes:
        value (float): Hbmin value.
    """

    def __init__(self, value: Union[str, float] = None):
        self.value: float = float(value) if value is not None else 0.9
        self._raise_exception_if_incorrect(self.value)

    def _raise_exception_if_incorrect(self, value: float) -> None:
        """Check if Hbmin value is correct, raise Exception if not.

        Args:
            value: Hbmin value to be checked.
        Raises:
            TypeError: If value is not float or not convertable to float.
            ValueError: If value is not in range [0,1].
        """

        try:
            v = float(value)
        except ValueError:
            raise TypeError(f"{value} was passed, a float excpected.")
        if v < 0 or v > 1:
            raise ValueError(f"Hbmin must be in range [0,1].")


class Stop(object):
    """Minimum value of AffinityTree leaves mincomp.

    Args:
        value: Stop value.

    Raises:
        ValueError: If Stop value is not in range [0,1].

    Attributes:
        value (float): The stop value.
    """

    def __init__(self, value: Union[str, float] = None):
        self.value: float = float(value) if value is not None else 0.99
        self._raise_exception_if_incorrect(self.value)

    def _raise_exception_if_incorrect(self, value: float) -> None:
        if value < 0:
            raise ValueError("STOP must be greater than 0.")
        if value > 1:
            raise ValueError("STOP must be smaller than 1.")


class P(object):
    """Parameter that changes compatiblity linear meaning to compatibility**P.

    Args:
        value: P value.

    Attributes:
        value (float): P value.
    """

    def __init__(self, value: Union[str, float] = None):
        self.value: float = float(value) if value is not None else 1



from typing import Optional, List, Tuple, Dict

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment

from pangtreebuild.tools import logprocess
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.pangenome import graph

_ParsedMaf = List[Optional[MultipleSeqAlignment]]

global_logger = logprocess.get_global_logger()
detailed_logger = logprocess.get_logger("details")


def get_poagraph(maf: msa.Maf, metadata: Optional[msa.MetadataCSV]) -> \
        Tuple[List[graph.Node], Dict[msa.SequenceID, graph.Sequence]]:
    """Get poagraph elements from MAF.

    Args:
        maf: Multialignment file in MAF format.
        metadata: MetadataCSV.

    Returns:
        Tuple of poagraph elements.
    """

    alignment = [*AlignIO.parse(maf.filecontent, "maf")]
    nodes, sequences = _init_poagraph(alignment, metadata)

    current_node_id = graph.NodeID(-1)
    column_id = graph.ColumnID(-1)
    for block_id, block in enumerate(alignment):
        global_logger.info(f"Processing block {block_id}...")
        block_width = len(block[0].seq)

        for col in range(block_width):
            column_id += 1
            sequence_id_to_nucleotide = {msa.SequenceID(seq.id): seq[col]
                                         for seq in block}
            nodes_codes = sorted([*(
                set([nucleotide
                     for nucleotide in sequence_id_to_nucleotide.values()])).difference({'-'})])
            column_nodes_ids = [graph.NodeID(current_node_id + i + 1)
                                for i, _ in enumerate(nodes_codes)]

            for i, nucl in enumerate(nodes_codes):
                current_node_id += 1
                nodes.append(graph.Node(node_id=current_node_id,
                                        base=graph.Base(nucl),
                                        aligned_to=_get_next_aligned_node_id(graph.NodeID(i), column_nodes_ids),
                                        column_id=graph.ColumnID(column_id),
                                        block_id=graph.BlockID(block_id)
                                        )
                             )

                for seq_id, nucleotide in sequence_id_to_nucleotide.items():
                    if nucleotide == nucl:
                        sequences[seq_id] = _add_node_do_sequence(sequence=sequences[seq_id],
                                                                  node_id=current_node_id)

    return nodes, sequences


def _init_poagraph(alignment: _ParsedMaf,
                   metadata: Optional[msa.MetadataCSV]) -> \
        Tuple[List[graph.Node], Dict[msa.SequenceID, graph.Sequence]]:
    maf_sequences_ids = _get_sequences_ids(alignment)
    metadata_sequences_ids = metadata.get_all_sequences_ids() if metadata else []
    initial_sequences = {seq_id: graph.Sequence(seqid=seq_id,
                                                paths=[],
                                                seqmetadata=metadata.get_sequence_metadata(seq_id)
                                                if metadata else {})
                         for seq_id in set(maf_sequences_ids + metadata_sequences_ids)}

    return [], initial_sequences


def _get_sequences_ids(alignment: _ParsedMaf) -> List[msa.SequenceID]:
    return list({msa.SequenceID(seq.id)
                 for block in alignment for seq in block})


def _get_next_aligned_node_id(current_column_i: graph.NodeID,
                              column_nodes_ids: List[graph.NodeID]) -> \
        Optional[graph.NodeID]:
    if len(column_nodes_ids) > 1:
        return column_nodes_ids[(current_column_i + 1) % len(column_nodes_ids)]
    return None


def _add_node_do_sequence(sequence: graph.Sequence, node_id: graph.NodeID) -> \
        graph.Sequence:
    if sequence.paths:
        a = graph.SeqPath([node_id])
        updated_path = graph.SeqPath(sequence.paths[-1] + a)
        newpaths = [sequence.paths[:-1] + updated_path]
    else:
        newpaths = [graph.SeqPath([node_id])]
    return graph.Sequence(sequence.seqid, newpaths, sequence.seqmetadata)



import os
import sys
import datetime


sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../pangtreebuild')))
from pangtreebuild.affinity_tree import builders as at_builders
from pangtreebuild.serialization import fasta
from pangtreebuild.serialization import json
from pangtreebuild.serialization import po
from pangtreebuild.pangenome import builder
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import cli, pathtools, logprocess


def main():
    parser = cli.get_parser()
    args = parser.parse_args()
    start = datetime.datetime.now()
    if not args.quiet and args.verbose:
        logprocess.add_file_handler_to_logger(args.output_dir, "details", "details.log", propagate=False)
        logprocess.add_file_handler_to_logger(args.output_dir, "", "details.log", propagate=False)
    if args.quiet:
        logprocess.disable_all_loggers()

    poagraph, dagmaf, fasta_provider = None, None, None
    if isinstance(args.multialignment, msa.Maf) and args.raw_maf:
        poagraph = builder.build_from_maf(args.multialignment, args.metadata)
    elif isinstance(args.multialignment, msa.Maf) and not args.raw_maf:
        fasta_provider = cli.resolve_fasta_provider(args)
        poagraph, dagmaf = builder.build_from_dagmaf(args.multialignment,
                                                     fasta_provider,
                                                     args.metadata)
    elif isinstance(args.multialignment, msa.Po):
        poagraph = builder.build_from_po(args.multialignment, args.metadata)

    affinity_tree = None
    if args.affinity is not None:
        blosum = args.blosum if args.blosum else cli.get_default_blosum()
        if fasta_provider is not None and isinstance(fasta_provider, missings.ConstBaseProvider):
            blosum.check_if_symbol_is_present(fasta_provider.missing_base.as_str())

        consensus_output_dir = pathtools.get_child_dir(args.output_dir, "affinitytree")

        if args.affinity == 'poa':
            affinity_tree = at_builders.build_poa_affinity_tree(poagraph,
                                                                blosum,
                                                                consensus_output_dir,
                                                                args.hbmin,
                                                                args.verbose)
        elif args.affinity == 'tree':
            affinity_tree = at_builders.build_affinity_tree(poagraph,
                                                            blosum,
                                                            consensus_output_dir,
                                                            args.stop,
                                                            args.p,
                                                            args.verbose)
        if args.metadata is not None:
            seq_id_to_metadata = {seq_id: seq.seqmetadata
                                  for seq_id, seq in poagraph.sequences.items()}
        else:
            seq_id_to_metadata = None

        affinity_tree_newick = affinity_tree.as_newick(seq_id_to_metadata,
                                                       separate_leaves=True)

        pathtools.save_to_file(affinity_tree_newick,
                               pathtools.get_child_path(consensus_output_dir, "affinity_tree.newick"))

    if args.output_po:
        pangenome_po = po.poagraph_to_PangenomePO(poagraph)
        pathtools.save_to_file(pangenome_po, pathtools.get_child_path(args.output_dir, "poagraph.po"))

    if args.output_fasta:
        sequences_fasta = fasta.poagraph_to_fasta(poagraph)
        pathtools.save_to_file(sequences_fasta, pathtools.get_child_path(args.output_dir, "_sequences.fasta"))
        if affinity_tree:
            consensuses_fasta = fasta.affinity_tree_to_fasta(poagraph, affinity_tree)
            pathtools.save_to_file(consensuses_fasta, pathtools.get_child_path(args.output_dir, "affinitytree.fasta"))

    end = datetime.datetime.now()
    pangenomejson = json.to_PangenomeJSON(task_parameters=cli.get_task_parameters(args, running_time=f"{end-start}s"),
                                          poagraph=poagraph,
                                          dagmaf=dagmaf,
                                          affinity_tree=affinity_tree)

    pangenome_json_str = json.to_json(pangenomejson)
    pathtools.save_to_file(pangenome_json_str, pathtools.get_child_path(args.output_dir, "pangenome.json"))


if __name__ == "__main__":
    main()



"""Data model of poagraph"""

from enum import Enum
import numpy as np
from typing import Any, Dict, List, NewType, Optional, Union

from pangtreebuild.affinity_tree import parameters as at_params
from pangtreebuild.pangenome.parameters import msa


class DataType(Enum):
    """It describes whether poagraph is build from nucleotides or proteins."""

    Nucleotides = 0
    Proteins = 1


class Base(object):
    """Nucleotide or protein base.

    Args:
        base: base value as str

    Raises:
        ValueError: If base has length other than 1.

    Attributes:
        value (bytes): base value
    """

    def __init__(self, base: str):
        if len(base) == 0:
            raise ValueError("Poagraph node base cannot be empty.")
        if len(base) > 1:
            raise ValueError("Poagraph node base must have length = 1.")
        self.value: bytes = str.encode(base)

    def __eq__(self, other: 'Base'):
        return other and self.value.__eq__(other.value)

    def as_str(self) -> str:
        """Returns base value as string."""

        return self.value.decode("ASCII")


NodeID = NewType('NodeID', int)
ColumnID = NewType('ColumnID', int)
BlockID = NewType('BlockID', int)


class Node(object):
    """Poagraph node.

    Args:
        node_id: Node ID.
        base: Node base (single letter).
        aligned_to: ID of the poagraph node it is aligned to.
        column_id: Poagraph is visualised as succeeding columns of nodes,
                   so nodes have assigned a (not unique) column ID.
        block_id: If multialignment source was MAF this is ID of the block that the node is from. Otherwise == 1.

    Attributes:
        node_id: Node ID.
        base: Node base (single letter).
        aligned_to: ID of the poagraph node it is aligned to.
        column_id: Poagraph is visualised as succeeding columns of nodes,
                   so nodes have assigned a (not unique) column ID.
        block_id: If multialignment source was MAF this is ID of the block that the node is from. Otherwise == 1.
    """
    def __init__(self,
                 node_id: NodeID,
                 base: Base,
                 aligned_to: Optional[NodeID] = None,
                 column_id: Optional[ColumnID] = None,
                 block_id: Optional[BlockID] = None):
        self.node_id: NodeID = node_id
        self._base: Base = base
        self.aligned_to: NodeID = aligned_to
        self.column_id: ColumnID = column_id
        self.block_id: BlockID = block_id

    def get_base(self) -> str:
        """
        Returns base the node represents.

        Returns: Node base as str.
        """

        return self._base.value.decode("ASCII")

    def __eq__(self, other: 'Node'):
        return (self.node_id == other.node_id
                and self._base == other._base
                and self.aligned_to == other.aligned_to
                )

    def __str__(self):
        return \
            f"id: {self.node_id}, " \
            f"base: {self.get_base()}, " \
            f"aligned_to: {self.aligned_to}, " \
            f"column_id: {self.column_id}, " \
            f"block_id: {self.block_id}"

    def __repr__(self):
        return self.__str__()


SeqPath = NewType('SeqPath', List[NodeID])
SequenceMetadata = NewType('Metadata', Dict[str, Any])


class Sequence(object):
    """Sequence present in pangenome defined as SequenceID, SeqPath (list of list of nodes ids) and metadata.

    Args:
        seqid: Sequence ID.
        paths: List of SeqPaths as the sequence can be splitted into several paths.
        seqmetadata: Metadata connected with this sequence.

    Attributes:
        seqid (msa.SequenceID): Sequence ID.
        paths (List[SeqPath]): List of SeqPaths as the sequence can be splitted into several paths.
        seqmetadata (SequenceMetadata): Dictionary of metadata connected with this sequence.
    """

    def __init__(self, seqid: msa.SequenceID, paths: List[SeqPath], seqmetadata: SequenceMetadata):
        self.seqid: msa.SequenceID = seqid
        self.paths: List[SeqPath] = paths
        self.seqmetadata: SequenceMetadata = seqmetadata

    def __str__(self):
        return f"seqid: {self.seqid}, paths: {[str(p) for p in self.paths]}, metadata: {str(self.seqmetadata)}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: 'Sequence'):
        return other and \
                self.seqid == other.seqid and \
                self.paths == other.paths and \
                self.seqmetadata == other.seqmetadata


class Compatibility(object):
    """Asymetric similiarity measure of two poagraph paths.

    Args:
        compatibility: Raw compatibility value - count of common nodes devided by length of one of the paths.
        p: Parameter to control compatibility value interpretation. Compatibility is raised to the power of P.

    Attributes:
        value (float): Compatibility value raised to the power of p.
        p (float): P parameter value.

    """

    def __init__(self, compatibility: float, p: at_params.P = at_params.P(1)):
        self.value: float = compatibility**p.value
        self.p: float = p.value

    def _check_p_equality(self, other: Union["Compatibility", Any]) -> None:
        if isinstance(other, Compatibility):
            assert self.p == other.p, 'Cannot compare compatibilities with different p values.'
        else:
            return

    def __eq__(self, other: Union["Compatibility", at_params.Stop]) -> bool:
        self._check_p_equality(other)
        return self.value == other.value

    def __lt__(self, other: Union["Compatibility", at_params.Stop]) -> bool:
        self._check_p_equality(other)
        return self.value < other.value

    def __le__(self, other: Union["Compatibility", at_params.Stop]) -> bool:
        self._check_p_equality(other)
        return self.value <= other.value

    def __gt__(self, other: Union["Compatibility", at_params.Stop]) -> bool:
        self._check_p_equality(other)
        return self.value > other.value

    def __ge__(self, other: Union["Compatibility", at_params.Stop]) -> bool:
        self._check_p_equality(other)
        return self.value >= other.value

    def __sub__(self, other: Union["Compatibility", at_params.Stop]) -> "Compatibility":
        self._check_p_equality(other)
        return Compatibility(self.value - other.value, at_params.P(self.p))

    def __str__(self) -> str:
        return f"""{self.value}"""

    def __repr__(self) -> str:
        return f"""value: {self.value}, p: {self.p}"""

    def base_value(self) -> "Compatibility":
        """Returns compatibility value without P transformation.

        Returns:
        Compatibility object with the original compatibility value and P=1."""

        return Compatibility(self.value ** (1 / self.p))


class Poagraph(object):
    """Pangenome is defined in Pangtree as poagraph which consists of connected nodes and sequences paths.

    Args:
        nodes: List of nodes.
        sequences: Dictonary of sequences IDs and corresponding sequences objects.
        datatype: Type of the data in poagraph.

    Attributes:
        nodes: List of nodes.
        sequences: Dictonary of sequences IDs and corresponding sequences objects.
        datatype: Type of the data in poagraph.

    """
    def __init__(self,
                 nodes: List[Node],
                 sequences: Dict[msa.SequenceID, Sequence],
                 datatype: Optional[DataType] = DataType.Nucleotides):
        self.nodes: List[Node] = nodes
        self.sequences: Dict[msa.SequenceID, Sequence] = sequences
        self.datatype: DataType = datatype

    def __eq__(self, other: 'Poagraph') -> bool:
        return self.nodes == other.nodes and \
            self.sequences == other.sequences and \
            self.datatype == other.datatype

    def get_compatibilities(self,
                            sequences_ids: List[msa.SequenceID],
                            consensus_path: SeqPath,
                            p: Optional[at_params.P] = at_params.P(1)) -> Dict[msa.SequenceID, Compatibility]:
        """Calculate compatibilities of sequences listed in sequences_ids to given consensus_path. Use P.

        Args:
            sequences_ids: Compatibilities of these seqeunces will be calculated.
            consensus_path: Sequences will be compared to this consensus path.
            p: Parameter P, see affinity tree algorithm parameters.

        Returns:
            Dictionary of sequences_ids and corresponding compatibility to given consensus_path.

        Raises:
            KeyError: If there is no sequence with given ID.
        """

        compatibilities = dict()
        for seq_id in sequences_ids:
            try:
                sequence_paths = self.sequences[seq_id].paths
            except KeyError:
                raise Exception("No sequence with given ID in poagraph.")
            if len(sequence_paths) == 1:
                sequence_path = sequence_paths[0]
            else:
                sequence_path = [node_id for path in sequence_paths for node_id in path]
            compatibilities[seq_id] = Compatibility(len(set(sequence_path).intersection(set(consensus_path))) /
                                                    len(sequence_path), p)
        return compatibilities

    def get_sequences_weights(self, sequences_ids: List[msa.SequenceID]) -> \
            Dict[msa.SequenceID, int]:
        """Calculates and normalizes sequences weights inside given group of sequences.

        Args:
            sequences_ids: IDs of the sequences that the weights must be calculated and normalized.

        Returns:
            Dictionary of sequences IDs and corresponding weights.
        """

        if not sequences_ids:
            return dict()

        a = np.zeros(len(self.nodes), dtype=np.int)
        unweighted_sources_weights = {}
        for seq_id in sequences_ids:
            for path in self.sequences[seq_id].paths:
                a[path] += 1

        for seq_id in sequences_ids:
            sequence_nodes_ids = [node_id for path in self.sequences[seq_id].paths for node_id in path]
            unweighted_sources_weights[seq_id] = np.mean(a[sequence_nodes_ids]) if any(a[sequence_nodes_ids]) else 0

        max_weight = max(unweighted_sources_weights.values())
        min_weight = min(unweighted_sources_weights.values())
        diff_weight = max_weight - min_weight
        if diff_weight == 0:
            normalized_sources_weights_dict = {path_key: 100 for path_key in unweighted_sources_weights.keys()}
        else:
            normalized_sources_weights_dict = {path: int((weight - min_weight)/diff_weight*100)
                                               for path, weight in unweighted_sources_weights.items()}
        # return {path_key: 100 for path_key in unweighted_sources_weights.keys()}
        return normalized_sources_weights_dict

    def get_sequence_nodes_count(self, seq_id: msa.SequenceID) -> int:
        """Returns length of indicated sequence (sum of lengths of all sequence paths).

        Args:
            seq_id: ID of the sequence.

        Returns:
            Sum of lengths of all sequence paths.
        """

        if seq_id not in self.sequences:
            raise Exception("No sequence with given ID in poagraph.")
        return sum([len(path) for path in self.sequences[seq_id].paths])

    def get_sequences_ids(self) -> List[msa.SequenceID]:
        """Returns _sequences of all genome _sequences in Poagraph.

        Returns:
            List of sequences IDs.
        """

        return [*self.sequences.keys()]

    @staticmethod
    def complement_metadata_for_sequences_absent_in_metadata_provided(poagraph: 'Poagraph',
                                                                      metadata: msa.MetadataCSV) -> None:
        """Complements metadata in given poagraph using provided metadata.

        Args:
            poagraph: The poagraph to be changed.
            metadata: Metadata to be inserted.
        """

        headers = metadata.get_metadata_keys()
        for seq in poagraph.sequences.values():
            for h in headers:
                if h not in seq.seqmetadata:
                    seq.seqmetadata[h] = None



# -*- coding: utf-8 -*-
from .Arc import Arc

class Block:
    def __init__(self, block_id, alignment):
        self.id = block_id
        self.alignment = alignment
        self.toroot = self # parent in find-union tree
        self.shift = 0 # order relative to parent
        self.reorder_shift = 0 # modification caused by edges inserted within group
        self.orient = 1 # orientation relative to parent
        self.flanks = {-1: 0, 1: 0} # blocks in group with negative/positive order (only in root)
        self.out_edges = []
        
    def find(self):
        if self.toroot is self:
            return self
        else:
            root = self.toroot.find()
            # let's update atributes for flattened tree structure
            self.orient *= self.toroot.orient
            self.shift = self.shift*self.toroot.orient+self.toroot.shift
            self.reorder_shift *= self.toroot.orient
            self.toroot = root
        return root
    
    def orientation(self):
        self.find()
        return self.orient

    def order(self):
        self.find()
        return self.shift+self.reorder_shift
    
    def reorder(self, n):
        self.reorder_shift += n - self.order()

#    def reorder(self, delta):
#        self.reorder_shift += delta

    def size(self):
        rootflanks = self.find().flanks
        return rootflanks[1]+rootflanks[-1]+1
    
    def minimum(self):
        root = self.find()
        return -root.flanks[-1]

    def maximum(self):
        root = self.find()
        return root.flanks[1]

    def unionto(self, other, reverse, flank): # join self to other
        selfroot = self.find()
        otheroot = other.find()
        selfroot.orient = reverse
        selfroot.reorder_shift *= reverse
        selfroot.shift = flank*(otheroot.flanks[flank]+selfroot.flanks[-reverse*flank]+1)
        otheroot.flanks[flank] += selfroot.flanks[-1]+selfroot.flanks[1]+1
        selfroot.toroot = otheroot
        del selfroot.flanks
        
    def orient_block(self):
        # Modify alignment according to block orientation
        if self.orientation() == -1:
            for u in self.alignment:
                u.seq = u.seq.reverse_complement()
                u.annotations["strand"] *= -1
                u.annotations["start"] = u.annotations["srcSize"] - u.annotations["size"] - u.annotations["start"]
        
    def add_out_edges(self, to, edgeType, list_of_seq):
        self.out_edges.append(Arc(to, edgeType, list_of_seq))



from typing import Optional, Tuple

from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.builders import maf2poagraph,\
                                             maf2dagmaf,\
                                             dagmaf2poagraph,\
                                             po2poagraph
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.pangenome import DAGMaf


def build_from_maf(maf: msa.Maf,
                   metadata: Optional[msa.MetadataCSV] = None,
                   datatype: Optional[graph.DataType] = graph.DataType.Nucleotides) -> \
        graph.Poagraph:
    """Builds poagraph from MAF file.

    Args:
        maf: Multialignment as MAF file.
        metadata: Metadata of sequences present in MAF.
        datatype: Type of the processed data (nucleotides/proteins).

    Returns:
        Poagraph based on given input data.
    """

    nodes, sequences = maf2poagraph.get_poagraph(maf, metadata)
    p = graph.Poagraph(nodes, sequences)
    if metadata:
        graph.Poagraph.complement_metadata_for_sequences_absent_in_metadata_provided(p, metadata)
    p.datatype = datatype
    return p


def build_from_dagmaf(maf: msa.Maf,
                      fasta_provider: Optional[missings.FastaProvider] = missings.ConstBaseProvider(
                          missings.MissingBase()),
                      metadata: Optional[msa.MetadataCSV] = None,
                      datatype: Optional[graph.DataType] = graph.DataType.Nucleotides) -> \
        Tuple[graph.Poagraph, DAGMaf.DAGMaf]:
    """Converts MAF to DagMaf and builds poagraph from MAF file.

    Args:
        maf: Multialignment as MAF file.
        fasta_provider: Provider of bases missing in DagMaf.
        metadata: Metadata of sequences present in MAF.
        datatype: Type of the processed data (nucleotides/proteins).

    Returns:
        Tuple: poagraph based on given input data and dagmaf created
            from input MAF.
    """

    dagmaf = maf2dagmaf.get_dagmaf(maf)
    nodes, sequences = dagmaf2poagraph.get_poagraph(dagmaf,
                                                    fasta_provider,
                                                    metadata)
    p = graph.Poagraph(nodes, sequences)
    if metadata:
        graph.Poagraph.complement_metadata_for_sequences_absent_in_metadata_provided(p, metadata)
    p.datatype = datatype
    return p, dagmaf


def build_from_po(po: msa.Po,
                  metadata: Optional[msa.MetadataCSV] = None,
                  datatype: Optional[graph.DataType] = graph.DataType.Nucleotides) -> graph.Poagraph:
    """Builds poagraph from PO file.

    Args:
        po: Multialignment as PO file.
        metadata: Metadata of sequences present in MAF.
        datatype: Type of the processed data (nucleotides/proteins).

    Returns:
        Poagraph based on given input data.
    """

    nodes, sequences = po2poagraph.get_poagraph(po, metadata)
    p = graph.Poagraph(nodes, sequences)
    if metadata:
        graph.Poagraph.complement_metadata_for_sequences_absent_in_metadata_provided(p, metadata)
    p.datatype = datatype
    return p



"""Affinity Tree builders"""

from collections import deque
import numpy as np
from pathlib import Path
from typing import Dict, List, Tuple, Optional

from pangtreebuild.tools import logprocess
from pangtreebuild.tools.cli import get_default_blosum
from pangtreebuild.affinity_tree import parameters
from pangtreebuild.affinity_tree import poa
from pangtreebuild.affinity_tree import tree

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa


global_logger = logprocess.get_global_logger()
tresholds_logger = logprocess.get_logger('tresholdsCSV')
detailed_logger = logprocess.get_logger('details')


class AffinityTreeBuildException(Exception):
    """Any exception connected with Affinity Tree build process."""

    pass


def build_poa_affinity_tree(p: graph.Poagraph,
                            blosum: Optional[parameters.Blosum],
                            output_dir: Path,
                            hbmin: parameters.Hbmin,
                            verbose: bool) -> tree.AffinityTree:
    """Builds Affinity Tree coherent with poa software.

    This method builds a simple version of Affinity Tree
    as it uses a single call to poa software. Poa provides
    division of sequences in Poagraph into consistent groups
    with a consensus path assigned to each group. These groups
    are converted in this method to Affinity Tree nodes and
    connected with a dummy root node so the result is coherent
    with pangtree definition of Affinity Tree.

    Args:
        p: Poagraph containing sequences to be divided into
            groups (Affinity Tree nodes).
        optional blosum: BLOSUM matrix. If not provided, default Blosum80.mat is used.
        output_dir: Path to a directory that can be used by poa software.
        hbmin: Parameter required by poa software. The minimum value of
            sequence compatibility to generated consensus.
        verbose: Switch to control logging intensity.

    Raises:
        AffinityTreeGenerationException: if consensuses cannot be found.
    """
    def _convert_consensus_paths_to_affinity_tree_nodes():
        at_nodes = []
        assigned_sequences = []
        for c_id, c_info in consensus_paths.items():
            assigned_sequences += c_info.assigned_sequences_ids
            all_seq = p.get_sequences_ids()
            compatibilities = p.get_compatibilities(all_seq, c_info.path)
            if len(c_info.assigned_sequences_ids):
                assigned_seq_comp = [c
                                     for seq_id, c in compatibilities.items()
                                     if seq_id in c_info.assigned_sequences_ids]
                mincomp = min(assigned_seq_comp)
            else:
                mincomp = 0
            new_node = tree.AffinityNode(id_=tree.AffinityNodeID(c_id + 1),
                                         parent=tree.AffinityNodeID(0),
                                         sequences=c_info.assigned_sequences_ids,
                                         mincomp=mincomp,
                                         compatibilities=compatibilities,
                                         consensus=c_info.path,
                                         children=[])
            at_nodes.append(new_node)

        node_for_unassigned_sequences = tree.AffinityNode(parent=tree.AffinityNodeID(0),
                                                          sequences=[seq_id
                                                                     for seq_id in p.get_sequences_ids()
                                                                     if seq_id not in assigned_sequences],
                                                          id_=tree.AffinityNodeID(len(at_nodes) + 1),
                                                          mincomp=p.Compatibility(0),
                                                          children=[])
        at_nodes.append(node_for_unassigned_sequences)
        return at_nodes

    global_logger.info("POA defined affinity tree generation started.")
    if blosum is None:
        blosum = get_default_blosum()
    _raise_error_if_invalid_poagraph(p)
    try:
        consensus_paths = poa.get_consensuses(p,
                                              p.get_sequences_ids(),
                                              output_dir,
                                              "poa_tree",
                                              blosum.filepath,
                                              hbmin)
    except poa.NoConsensusError:
        raise AffinityTreeBuildException("No consensus in the Affinity Tree.")

    consensus_nodes = _convert_consensus_paths_to_affinity_tree_nodes()
    root_node = tree.AffinityNode(id_=tree.AffinityNodeID(0),
                                  children=[c_node.id_
                                            for c_node in consensus_nodes])
    affinity_tree = tree.AffinityTree([root_node] + consensus_nodes)
    global_logger.info("POA defined affinity tree generation started.")
    return affinity_tree


def build_affinity_tree(poagraph: graph.Poagraph,
                        blosum: Optional[parameters.Blosum],
                        output_dir: Path,
                        stop: parameters.Stop,
                        p: parameters.P,
                        verbose: bool) -> tree.AffinityTree:
    """Builds Affinity Tree.

    Affinity Tree is defined in paper 'Getting insight into the
    pan-genome structure with Pangtree'. This method builds
    an Affinity Tree by iterative calls to poa software.
    Full algorithm and idea are described in the above-mentioned paper.

    Args:
        poagraph: Poagraph containing _sequences to be divided into groups
            (Affinity Tree nodes).
        optional blosum: BLOSUM matrix. If not provided, default Blosum80.mat is used.
        output_dir: Path to a directory that can be used by poa software.
        stop: Value of mincomp above which an affinity tree node is no more
            split.
        p: Value changing the linear meaning of compatibility when searching
            for cutoff.
        verbose: Switch to control logging intensity.

    Raises:
        AffinityTreeGenerationException: if consensuses cannot be found.

    Returns:
        Affinity Tree generated with Pangtree algorithm.
    """

    global_logger.info("Affinity Tree generation started.")
    if blosum is None:
        blosum = get_default_blosum()
    if verbose:
        logprocess.add_file_handler_to_logger(output_dir,
                                              "tresholdsCSV",
                                              "tresholds.csv",
                                              "%(message)s", False)
    _raise_error_if_invalid_poagraph(poagraph)

    root_node = _get_root_node(poagraph, blosum.filepath, output_dir, p)
    affinity_tree = tree.AffinityTree([root_node])

    nodes_to_process = deque([affinity_tree.get_node(tree.AffinityNodeID(0))])
    while nodes_to_process:
        node = nodes_to_process.pop()

        children_nodes = _get_children_nodes_looping(node,
                                                     poagraph,
                                                     output_dir,
                                                     blosum.filepath,
                                                     p,
                                                     affinity_tree.get_max_node_id())
        if len(children_nodes) == 1:
            continue

        for child in children_nodes:
            all_sequences = [*poagraph.sequences.keys()]
            child.compatibilities = poagraph.get_compatibilities(sequences_ids=all_sequences,
                                                                 consensus_path=child.consensus,
                                                                 p=p)
            node.children.append(child.id_)
            affinity_tree.nodes.append(child)
            if not _node_is_ready(child, stop):
                nodes_to_process.append(child)
    global_logger.info("Affinity Tree generation finished.\n")
    return affinity_tree


def _raise_error_if_invalid_poagraph(poagraph: graph.Poagraph) -> None:
    """Checks if any sequence is present in the input poagraph.

    Args:
        poagraph: Affinity Tree generation algorithm input poagraph.
    Raises:
        AffinityTreeGenerationException: If no sequence is present
            in the poagraph.
    """

    if len(poagraph.sequences) == 0:
        raise AffinityTreeBuildException("""Invalid poagraph. No paths in
                                            poagraph. Impossible Affinity
                                            Tree generation.""")


def _get_root_node(poagraph: graph.Poagraph,
                   blosum_path: Path,
                   output_dir: Path,
                   p: parameters.P) -> tree.AffinityNode:
    """Creates root node of the Affinity Tree.

    The node has assigned a consensus path and
    all sequences present in poagraph."""

    detailed_logger.info("Getting the root affinity node...")
    all_poagraph_sequences_ids = poagraph.get_sequences_ids()
    try:
        consensus_paths = poa.get_consensuses(poagraph,
                                              all_poagraph_sequences_ids,
                                              output_dir,
                                              "root",
                                              blosum_path,
                                              hbmin=parameters.Hbmin(0),
                                              specific_consensuses_id=[0])
    except poa.NoConsensusError:
        raise AffinityTreeBuildException("Cannot find root consensus.")
    compatibilities = poagraph.get_compatibilities(all_poagraph_sequences_ids,
                                                   consensus_paths[0].path,
                                                   p=p)
    affinity_node = tree.AffinityNode(id_=tree.AffinityNodeID(0),
                                      sequences=[*poagraph.sequences.keys()],
                                      mincomp=_get_min_comp(all_poagraph_sequences_ids, compatibilities),
                                      compatibilities=compatibilities,
                                      consensus=consensus_paths[0].path)
    detailed_logger.info(f"New affinity node created: {str(affinity_node)}")
    return affinity_node


def _get_children_nodes_looping(node: tree.AffinityNode,
                                poagraph: graph.Poagraph,
                                output_dir: Path,
                                blosum_path: Path,
                                p: parameters.P,
                                current_max_affinity_node_id: int) -> List[tree.AffinityNode]:
    """Generates children of given Affinity Tree node."""

    children_nodes: List[tree.AffinityNode] = []
    not_assigned_sequences_ids: List[msa.SequenceID] = node.sequences
    detailed_logger.info(f"""Getting children nodes for
                             affinity node {node.id_}...""")

    affinity_node_id = 0
    so_far_cutoffs: List[poagraph.Compatibility] = []
    while not_assigned_sequences_ids:
        detailed_logger.info(f"### Getting child {len(so_far_cutoffs)}...")
        child_ready = False
        attempt = 0
        current_candidates = not_assigned_sequences_ids
        while not child_ready:
            consensus_candidate = poa.get_consensuses(poagraph,
                                                      current_candidates,
                                                      output_dir,
                                                      f"parent_{node.id_}_child_{len(so_far_cutoffs)}_attempt_{attempt}",
                                                      blosum_path,
                                                      parameters.Hbmin(0),
                                                      specific_consensuses_id=[0])[0].path
            compatibilities_to_consensus_candidate = poagraph.get_compatibilities(sequences_ids=not_assigned_sequences_ids,
                                                                                  consensus_path=consensus_candidate,
                                                                                  p=p)
            compatibilities_to_consensus_candidate[msa.SequenceID("parent")] = node.mincomp
            qualified_sequences_ids_candidates, cutoff = _get_qualified_sequences_ids_and_cutoff(
                compatibilities_to_max_c=compatibilities_to_consensus_candidate,
                so_far_cutoffs=so_far_cutoffs,
                splitted_node_id=node.id_)

            if qualified_sequences_ids_candidates == current_candidates or attempt == 10:
                if attempt == 10:
                    detailed_logger.info("Attempt treshold 10 exceeded!")
                affinity_node_id += 1

                affinity_node = tree.AffinityNode(
                    id_=tree.AffinityNodeID(current_max_affinity_node_id + affinity_node_id),
                    parent=node.id_,
                    sequences=qualified_sequences_ids_candidates,
                    mincomp=_get_min_comp(node_sequences_ids=qualified_sequences_ids_candidates,
                                          comps_to_consensus=compatibilities_to_consensus_candidate),
                    consensus=graph.SeqPath(consensus_candidate))
                children_nodes.append(affinity_node)
                not_assigned_sequences_ids = list(set(not_assigned_sequences_ids) - set(qualified_sequences_ids_candidates))
                child_ready = True
                so_far_cutoffs.append(affinity_node.mincomp)
            else:
                current_candidates = qualified_sequences_ids_candidates
                attempt += 1

    detailed_logger.info("Children nodes generated.")

    return children_nodes


def _get_min_comp(node_sequences_ids: List[msa.SequenceID],
                  comps_to_consensus: Dict[msa.SequenceID,
                                           graph.Compatibility]) -> graph.Compatibility:
    """Find minimum compatibility from the compatibilities of given _sequences.

    Args:
        node_sequences_ids: List of sequences IDs that the search
            is limited to.
        comps_to_consensus: Source of compatibilities values.

    Returns:
        The minimum compatibility.

    Raises:
        AffinityTreeBuildException: If sequence IDs from the list can be found
            in the compatibilities dictionary.
    """

    compatibilities_of_node_sequences = [comp
                                         for seq_id, comp
                                         in comps_to_consensus.items()
                                         if seq_id in node_sequences_ids]
    if not compatibilities_of_node_sequences:
        raise AffinityTreeBuildException("""No mincomp because no sequences are
                                            assigned to this affinity node.""")
    return min(compatibilities_of_node_sequences)


def _get_qualified_sequences_ids_and_cutoff(compatibilities_to_max_c: Dict[msa.SequenceID, graph.Compatibility],
                                            so_far_cutoffs: List[graph.Compatibility], splitted_node_id) \
        -> Tuple[List[msa.SequenceID], graph.Compatibility]:
    """Choose sequences qualified to be enclosed in single node.

    The choice is based on their compatibilities values."""

    node_cutoff = _find_node_cutoff(compatibilities=[*compatibilities_to_max_c.values()],
                                    so_far_cutoffs=so_far_cutoffs)
    tresholds_logger.info(f"Splitting {splitted_node_id}; NODE; {compatibilities_to_max_c}; "
                          f"{node_cutoff.cutoff}; {node_cutoff.explanation}")
    compatibtle_sequences_ids = _get_sequences_ids_above_cutoff(compatibilities_to_max_c, node_cutoff.cutoff)
    detailed_logger.info(f"{len(compatibtle_sequences_ids)} _sequences ({compatibtle_sequences_ids} are "
                         f"qualified to be enclosed in this node, mincomp = {node_cutoff.cutoff}.")
    return compatibtle_sequences_ids, node_cutoff.cutoff


def _node_is_ready(node: tree.AffinityNode, stop: parameters.Stop) -> bool:
    """Checks if the node should be further split.

    The decision is based on Stop parameter and count of sequences assigned.

    Args:
        node: Node to be checked.
        stop: Stop parameter

    Returns:
        bool: Information whether the node should be further split.
    """

    if len(node.sequences) == 1 or node.mincomp.base_value() >= stop:
        detailed_logger.info(f"""Node {node.id_} satisfied requirements
                                 and won't be split!""")
        return True
    return False


class FindCutoffResult:
    """Result of cutoff finding.

    Args:
        cutoff: The minimum accepted Compatibility value.
        explanation: The explanation of the choice.

    Attributes:
        cutoff (Compatibility): The minimum accepted Compatibility value.
        explanation (str): The explanation of the choice.
    """
    def __init__(self, cutoff: graph.Compatibility, explanation: str):
        self.cutoff: graph.Compatibility = cutoff
        self.explanation: str = explanation


def _find_node_cutoff(compatibilities: List[graph.Compatibility],
                      so_far_cutoffs: List[graph.Compatibility]) -> FindCutoffResult:
    """Searches for cutoff.

    Args:
        compatibilities: List of compatibilities to be searched.
        so_far_cutoffs: Cutoffs from previous children from node
            being currently split.

    Returns:
         The cutoff and the explanationo as _FindCutoffResult.
    """

    if not so_far_cutoffs:
        cutoff = _find_max_distance(compatibilities)
        reason = "No so far cutoffs. Find max distance in sorted values."
    else:
        guard = min(so_far_cutoffs)
        sorted_comp = sorted(compatibilities)
        if guard <= sorted_comp[0]:
            cutoff = sorted_comp[0]
            reason = "guard < min(compatibilities). Return min(compatibilities)."
        elif guard >= sorted_comp[-1]:
            cutoff = _find_max_distance(compatibilities)
            reason = "guard > max(compatibilities). Find max distance in sorted values."
        else:
            first_comp_greater_than_guard_index = [i for i, c in enumerate(sorted_comp) if c > guard][0]
            cutoff = _find_max_distance(sorted_comp[0:first_comp_greater_than_guard_index + 1])
            reason = "Find max distance in sorted_comp[0:first_comp_greater_than_guard_index + 1]"
    return FindCutoffResult(cutoff, reason)


def _find_max_distance(compatibilities: List[graph.Compatibility]) -> graph.Compatibility:
    """Sorts compatibilities and searches for the largest gap.

    Args:
        compatibilities: List of compatibilities to be searched
            for the distance.

    Returns:
        The first compatibility value after the largest gap
            in sorted compatibilities.

    Raises:
        ValueError: If provided compatibilities list is empty.
    """

    def break_if_empty(compatibilities: List[graph.Compatibility]) -> None:
        if not list(compatibilities):
            raise ValueError("Empty compatibilities list. Cannot find cutoff.")

    break_if_empty(compatibilities)
    if len(compatibilities) == 1:
        return compatibilities[0]

    sorted_values = sorted(compatibilities)
    distances = np.array([sorted_values[i + 1] - sorted_values[i] for i in range(len(sorted_values) - 1)])
    max_distance_index: int = np.argmax(distances)
    return sorted_values[max_distance_index + 1]


def _get_sequences_ids_above_cutoff(compatibilities: Dict[msa.SequenceID, graph.Compatibility],
                                    cutoff: graph.Compatibility) -> List[msa.SequenceID]:
    """Returns Seuqences IDs with compatibility above required treshold.

    Args:
        compatibilities: Dictionary of _sequences IDs and corresponding
            compatibilities.
        cutoff: Required minimum compatibility value of the qualified
            sequences.

    Returns:
        List of Sequences IDs that have greater compatibility than required
            cutoff value.
    """
    return [seq_id
            for seq_id, comp in compatibilities.items()
            if comp >= cutoff and seq_id != msa.SequenceID("parent")]



import unittest
from pathlib import Path

from ddt import unpack, data, ddt

from pangtreebuild.pangenome import graph, builder
from pangtreebuild.pangenome.parameters import missings, msa
from pangtreebuild.tools import pathtools

def sid(x): return msa.SequenceID(x)


def sm(x): return graph.SequenceMetadata(x)


@ddt
class MetadataCSVTests(unittest.TestCase):

    def setUp(self) -> None:
        self.csv_files_dir = Path(__file__).parent.joinpath("csv_files").resolve()
        self.alignment_files_dir = Path(__file__).parent.joinpath("alignment_files").resolve()
        self.fasta_provider = missings.ConstBaseProvider(missings.MissingBase())

    def test_1_correct(self):
        metadata_path = self.csv_files_dir.joinpath("test_1_correct.csv")
        csv_content = pathtools.get_file_content_stringio(metadata_path)

        expected_metadata = {msa.SequenceID('s1'): {'name': 'sequence1', 'group': 'A'},
                             msa.SequenceID('s2'): {'name': 'sequence2', 'group': 'B'},
                             msa.SequenceID('s3'): {'name': 'sequence3', 'group': 'B'}
                             }
        m = msa.MetadataCSV(csv_content, metadata_path)
        actual_metadata = m.metadata

        self.assertEqual(expected_metadata, actual_metadata)

    def test_2_no_seqid(self):
        csv_path = self.csv_files_dir.joinpath("test_2_no_seqid.csv")
        csv_content = pathtools.get_file_content_stringio(csv_path)
        with self.assertRaises(Exception) as err:
            _ = msa.MetadataCSV(csv_content, csv_path)
        self.assertEqual(f"No \'seqid\' column in metadata csv.", str(err.exception))

    def test_3_empty_file(self):
        csv_path = self.csv_files_dir.joinpath("test_3_empty_file.csv")

        csv_content = pathtools.get_file_content_stringio(csv_path)
        with self.assertRaises(Exception) as err:
            _ = msa.MetadataCSV(csv_content, csv_path)
        self.assertEqual(f"Empty csv file.", str(err.exception))

    def test_4_seqid_is_last(self):
        metadata_path = self.csv_files_dir.joinpath("test_4_seqid_is_last.csv")
        csv_content = pathtools.get_file_content_stringio(metadata_path)

        expected_metadata = {msa.SequenceID('s1'): {'name': 'sequence1', 'group': 'A'},
                             msa.SequenceID('s2'): {'name': 'sequence2', 'group': 'B'},
                             msa.SequenceID('s3'): {'name': 'sequence3', 'group': 'B'}
                             }
        m = msa.MetadataCSV(csv_content, metadata_path)
        actual_metadata = m.metadata

        self.assertEqual(expected_metadata, actual_metadata)

    def test_5_double_seqid(self):
        csv_path = self.csv_files_dir.joinpath("test_5_double_seqid.csv")

        csv_content = pathtools.get_file_content_stringio(csv_path)
        with self.assertRaises(Exception) as err:
            _ = msa.MetadataCSV(csv_content, csv_path)
        self.assertEqual("Only one \'seqid\' column in metadata csv is allowed.", str(err.exception))

    def test_6_incorrect_commas_number(self):
        csv_path = self.csv_files_dir.joinpath("test_6_incorrect_commas_number.csv")

        csv_content = pathtools.get_file_content_stringio(csv_path)
        with self.assertRaises(Exception) as err:
            _ = msa.MetadataCSV(csv_content, csv_path)
        self.assertEqual("CSV metadata error. Different number of columns in line 0 than in header line.",
                         str(err.exception))

    def test_7_not_unique_seqids(self):
        csv_path = self.csv_files_dir.joinpath("test_7_not_unique_seqids.csv")

        csv_content = pathtools.get_file_content_stringio(csv_path)
        with self.assertRaises(Exception) as err:
            _ = msa.MetadataCSV(csv_content, csv_path)
        self.assertEqual("Repeated values in seqid column in metadata file. Make them unique.", str(err.exception))

    def test_8_correct_with_dots(self):
        metadata_path = self.csv_files_dir.joinpath("test_8_seqids_with_dots.csv")
        csv_content = pathtools.get_file_content_stringio(metadata_path)

        expected_metadata = {msa.SequenceID('s1'): {'name': 'sequence1', 'group': 'A'},
                             msa.SequenceID('s2'): {'name': 'sequence2', 'group': 'B'},
                             msa.SequenceID('s3'): {'name': 'sequence3', 'group': 'B'}
                             }
        m = msa.MetadataCSV(csv_content, metadata_path)
        actual_metadata = m.metadata

        self.assertEqual(expected_metadata, actual_metadata)

    def test_9_get_seqids(self):
        metadata_path = self.csv_files_dir.joinpath("test_1_correct.csv")
        csv_content = pathtools.get_file_content_stringio(metadata_path)

        expected_seqids = [msa.SequenceID('s1'),
                           msa.SequenceID('s2'),
                           msa.SequenceID('s3')]

        m = msa.MetadataCSV(csv_content, metadata_path)
        actual_seqids = m.get_all_sequences_ids()

        self.assertEqual(expected_seqids, actual_seqids)

    @data(("test_1_the_same_seqid_in_maf_as_in_csv_1",
           "test_10_1.maf", "test_1_correct.csv", "test_10_1.po", {
            sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'})}),

          ("test_2_seqid_in_maf_as_in_csv_1_but_with_single_dots",
           "test_10_2.maf", "test_1_correct.csv", "test_10_2.po",
           {sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'})}),

          ("test_3_seqid_in_maf_as_in_csv_1_but_with_double_dots",
           "test_10_3.maf", "test_8_seqids_with_dots.csv", "test_10_3.po",
           {sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'})}),

          ("test_4_more_seqids_in_maf_than_in_csv",
           "test_10_4.maf", "test_1_correct.csv", "test_10_4.po",
           {sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'}),
            sid('s4'): sm({'group': None, 'name': None})}),

          ("test_5_more_seqids_in_csv_than_in_maf",
           "test_10_5.maf", "test_1_correct.csv", "test_10_5.po",
           {sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'})}),

          ("test_6_seqids_in_csv_and_maf_all_differ",
           "test_10_6.maf", "test_1_correct.csv", "test_10_6.po",
           {sid('s1'): sm({'group': 'A', 'name': 'sequence1'}),
            sid('s2'): sm({'group': 'B', 'name': 'sequence2'}),
            sid('s3'): sm({'group': 'B', 'name': 'sequence3'}),
            sid('ss1'): sm({'group': None, 'name': None}),
            sid('ss2'): sm({'group': None, 'name': None})})
          )
    @unpack
    def test_10_metadata_feed_to_alignment_from_csv(self,
                                                    test_name,
                                                    maf_name,
                                                    csv_name,
                                                    po_name,
                                                    expected_metadata):
        maf_path = self.alignment_files_dir.joinpath(maf_name)
        csv_path = self.csv_files_dir.joinpath(csv_name)
        po_path = self.alignment_files_dir.joinpath(po_name)

        poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            msa.MetadataCSV(pathtools.get_file_content_stringio(csv_path), csv_path))
        actual_metadata = {seq_id: seq.seqmetadata
                           for seq_id, seq in poagraph.sequences.items()}
        self.assertEqual(expected_metadata, actual_metadata)

        poagraph = builder.build_from_maf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            msa.MetadataCSV(pathtools.get_file_content_stringio(csv_path), csv_path))
        actual_metadata = {seq_id: seq.seqmetadata
                           for seq_id, seq in poagraph.sequences.items()}
        self.assertEqual(expected_metadata, actual_metadata)

        poagraph = builder.build_from_po(
            msa.Po(pathtools.get_file_content_stringio(po_path), maf_path),
            msa.MetadataCSV(pathtools.get_file_content_stringio(csv_path), csv_path))
        actual_metadata = {seq_id: seq.seqmetadata
                           for seq_id, seq in poagraph.sequences.items()}
        self.assertEqual(expected_metadata, actual_metadata)


if __name__ == '__main__':
    unittest.main()



"""Functions to call poa software."""

from bisect import bisect_left
import os
from pathlib import Path
import subprocess
from typing import List, Dict, Union, Optional

from pangtreebuild.affinity_tree import parameters
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.serialization import po
from pangtreebuild.tools import pathtools
from pangtreebuild.tools import logprocess


detailed_logger = logprocess.get_logger('details')
global_logger = logprocess.get_global_logger()


class NoConsensusError(Exception):
    """If poa software cannot find any consensus path in the Poagraph"""

    pass


class ConsInfo(object):
    """Consensus information based on poa result.

    Args:
        fullname: Consensus path name. Eg. "CONSENS0"
        po_consensus_id: Consensus ID used in PO file to indicate it is
            present in a node. Eg. "S0"
        assigned_sequences_ids: IDs of _sequences assigned to this consensus.
        path: List of nodes of this consensus.

    Attributes:
        fullname (str): Consensus name.
        po_consensus_id (str): Consensus ID used in PO file to indicate
            it is present in a node. Eg. "S0"
        assigned_sequences_ids (List[msa.SequenceID]): IDs of sequences
            assigned to this consensus.
        path (graph.SeqPath): List of nodes of this consensus.
    """

    def __init__(self,
                 fullname: str,
                 po_consensus_id: Optional[str] = None,
                 assigned_sequences_ids: Optional[List[msa.SequenceID]] = None,
                 path: Optional[graph.SeqPath] = None):
        self.fullname: str = fullname
        self.po_consensus_id: str = po_consensus_id
        self.assigned_sequences_ids: List[msa.SequenceID] = assigned_sequences_ids
        self.path: graph.SeqPath = path


def get_consensuses(poagraph: graph.Poagraph,
                    sequences_ids: List[msa.SequenceID],
                    output_dir: Path,
                    job_name: str,
                    blosum_path: Path,
                    hbmin: parameters.Hbmin,
                    specific_consensuses_id: Optional[List[int]] = None) -> \
                        Dict[int, ConsInfo]:
    """Calls poa software on given Poagraph to get consensus paths.

    Args:
        poagraph: Poagraph used as input to poa software. It may be cropped by
            using sequences_ids argument.
        sequences_ids: IDs of the _sequences that should be kept in poagraph
            being input to poa.
        output_dir: Full path to the directory used by poa software as
            temporary storage place.
        job_name: Name of the task used to label produced file names.
        blosum_path: Full path to the Blosum file used as poa's input.
        hbmin: Hbmin value used used as poa's input.
        specific_consensuses_id: Poa returns consensuses numbered by: 0, 1...
            By this argument it can be specified which should be returned.

    Returns:
        Dictionary of consensus numbers and corresponding information as
            ConsInfo object.

    Raises:
        NoConsensusError: If no consensus was found for given Poagraph and
            set of selected _sequences.
    """
    poa_input_path = pathtools.get_child_path(output_dir,
                                              f"{job_name}_in_pangenome.po")
    poa_output_path = pathtools.get_child_path(output_dir,
                                               f"{job_name}_out_pangenome.po")

    s = _PoagraphPOTranslator(poagraph, sequences_ids)
    poa_input_content = s.get_input_po_content()
    with open(poa_input_path, 'w') as poa_input:
        poa_input.write(poa_input_content)
    _call_poa(po_file_path=poa_input_path,
              hb_file_path=poa_output_path,
              blosum_path=blosum_path.resolve(),
              hbmin=hbmin.value)
    with open(poa_output_path) as poa_output:
        poa_output_lines = poa_output.readlines()
    os.remove(poa_input_path)
    os.remove(poa_output_path)
    consensus_paths = s.read_consensus_paths(poa_output_lines,
                                             specific_consensuses_id)
    return consensus_paths


def _call_poa(po_file_path: Path,
              hb_file_path: Path,
              blosum_path: Path,
              hbmin: float) -> None:
    """Calls poa software.

    Args:
        po_file_path: Path to the PO file containing Poagraph.
        hb_file_path: Path to the PO file where the poa result will be saved.
        blosum_path: Path to the Blosum file.
        hbmin: Hbmin value used as poa input.

    Returns:
        Nothing, as the result is stored in hb_file_path.
    """
    affinity_tree_dir = Path(__file__).parent
    poa_path = pathtools.get_child_path(affinity_tree_dir, "bin/poa")
    detailed_logger.info(f"Run poa! Input: {po_file_path} Output: {hb_file_path}...")
    command = f"{poa_path} -read_msa {po_file_path} -hb -po {hb_file_path} {blosum_path} -hbmin {hbmin}"
    poa_result = subprocess.run(command, stderr=subprocess.PIPE, shell=True)
    poa_str_output = poa_result.stderr.decode("ASCII")
    detailed_logger.info(f"Poa output: {poa_str_output}")


class _PoagraphPOTranslator:
    """Converts Poagraph to PO file and back poa result PO file to Poagraph.

    Args:
        poagraph: Full poagraph where the consensuses are searched.
        sequences_ids: Sequences to be kept in the final PO file
            which stores Poagraph.

    Attributes:
        poagraph (poagraph.Poagraph): The orginal poagraph.
        sequences_ids (List[msa.SequenceID]): List of _sequences IDs to be
            included in poa input.
        new_to_old: Dict[poagraph.NodeID, poagraph.NodeID]: Mapping of
            temporary poagraph nodes IDs to the original ones.
        old_to_new: Dict[poagraph.NodeID, poagraph.NodeID] = Mapping of the
            original poagraph nodes IDs to the temporary ones.
        seq_old_to_new: Dict[msa.SequenceID, int] = Mapping of the original
            sequences IDs to temporary integer ones.
        seq_new_to_old: Dict[int, msa.SequenceID] = Mapping of the temporary
            int IDs of _sequences to the original ones.
    """
    def __init__(self,
                 poagraph: graph.Poagraph,
                 sequences_ids: List[msa.SequenceID]):
        self.poagraph: graph.Poagraph = poagraph
        self.sequences_ids: List[msa.SequenceID] = sequences_ids
        self.new_to_old: Dict[graph.NodeID, graph.NodeID] = None
        self.old_to_new: Dict[graph.NodeID, graph.NodeID] = None
        self.seq_old_to_new: Dict[msa.SequenceID, int] = None
        self.seq_new_to_old: Dict[int, msa.SequenceID] = None

    def get_input_po_content(self) -> str:
        """Convert poagraph to PO file content."""

        paths_to_keep = [path
                         for seq_id in self.sequences_ids
                         for path in self.poagraph.sequences[seq_id].paths]
        nodes_ids_to_keep = list(set([node_id
                                      for path in paths_to_keep
                                      for node_id in path]))
        sorted_nodes_ids_to_keep = sorted(nodes_ids_to_keep)
        self.old_to_new = {node_id: i
                           for i, node_id in enumerate(sorted_nodes_ids_to_keep)}
        self.new_to_old = {new_node_id: old_node_id
                           for old_node_id, new_node_id in self.old_to_new.items()}
        self.seq_old_to_new = {seq_id: i
                               for i, seq_id in enumerate(self.sequences_ids)}
        self.seq_new_to_old = {i: seq_id
                               for seq_id, i in self.seq_old_to_new.items()}

        po_nodes = [po.NodePO(base=self.poagraph.nodes[self.new_to_old[new_node_id]]._base.value,
                              aligned_to=self._get_aligned_node(self.new_to_old[new_node_id],
                                                                sorted_nodes_ids_to_keep),
                              in_nodes=set(),
                              sequences_ids=[])
                    for new_node_id in range(len(nodes_ids_to_keep))]

        sequences_weight: Dict[msa.SequenceID, int] = self.poagraph.get_sequences_weights(self.sequences_ids)
        po_sequences = [po.SequencePO(name=self.seq_new_to_old[new_seq_id],
                                      nodes_count=self.poagraph.get_sequence_nodes_count(self.seq_new_to_old[new_seq_id]),
                                      weight=sequences_weight[self.seq_new_to_old[new_seq_id]],
                                      consensus_id=-1,
                                      start_node_id=self.old_to_new[
                                          self.poagraph.sequences[self.seq_new_to_old[new_seq_id]].paths[0][0]])
                        for new_seq_id in range(len(self.sequences_ids))]

        for seq_id in self.sequences_ids:
            new_seq_id = self.seq_old_to_new[seq_id]
            for path in self.poagraph.sequences[seq_id].paths:
                for i, node_id in enumerate(path):
                    new_node_id = self.old_to_new[node_id]
                    po_nodes[new_node_id].sequences_ids.append(new_seq_id)
                    if i > 0:
                        new_in_node_id = self.old_to_new[path[i-1]]
                        po_nodes[new_node_id].in_nodes.add(new_in_node_id)

        for node in po_nodes:
            node.in_nodes = sorted(node.in_nodes)

        return po.poagraph_elements_to_PangenomePO(po_nodes,
                                                   po_sequences,
                                                   self.poagraph.datatype)

    def _get_aligned_node(self,
                          old_node_id: graph.NodeID,
                          sorted_nodes_ids_to_keep: List[graph.NodeID]) -> \
            Union[graph.NodeID, None]:
        """Get aligned node ID or None if it there is no aligned node.

        Args:
            old_node_id: Node ID in the original poagraph.
            sorted_nodes_ids_to_keep: Node IDs that are kept in the
                cropped Poagraph.

        Returns:
            Aligned node ID if it is present or None if not.
        """
        aligned_to = self.poagraph.nodes[old_node_id].aligned_to
        if aligned_to is None:
            return None
        while aligned_to != old_node_id:
            if self._is_in(sorted_list=sorted_nodes_ids_to_keep, x=aligned_to):
                return self.old_to_new[aligned_to]
            aligned_to = self.poagraph.nodes[aligned_to].aligned_to
        return None

    def _is_in(self, sorted_list: List[int], x: int) -> bool:
        """Checks if x is in sorted list.

        Args:
            sorted_list: List of sorted numerical values.
            x: Numerical value that is searched in sorted_list.

        Returns:
            Information whether x is in sorted_list or not.
        """

        i = bisect_left(sorted_list, x)
        if i != len(sorted_list) and sorted_list[i] == x:
            return True
        return False

    @staticmethod
    def _extract_line_value(line: str) -> str:
        """Specific to PO format. Extracts value assigned to variable in line.

        Args:
            Line of the PO file in format [...]=[...]

        Returns:
            Assigned value as str."""
        return line.split('=')[1].strip()

    def read_consensus_paths(self,
                             po_lines: List[str],
                             specific_consensuses_ids: Optional[List[int]] = None) -> Dict[int, ConsInfo]:
        """For given lines of PO file reads consensus information.

        Args:
            po_lines: Lines of the PO file that is poa result.
            specific_consensuses_ids: Poa generates as many consensuses as is possible but the number of read
                                      consensuses can be modifed by specifing IDs of consensuses to return.

        Returns:
            Dictionary of consensuses IDs and corresponding Consensus Info objects.
        """
        po_lines_iterator = iter(po_lines)

        for i in range(3):
            next(po_lines_iterator)

        nodes_count = int(_PoagraphPOTranslator._extract_line_value(next(po_lines_iterator)))
        paths_count = int(_PoagraphPOTranslator._extract_line_value(next(po_lines_iterator)))

        consensuses_in_po_lines: Dict[int, ConsInfo] = dict()
        current_consensus_id = -1
        for i in range(paths_count):
            path_name = _PoagraphPOTranslator._extract_line_value(next(po_lines_iterator))
            if len(path_name) > 7 and path_name[:7] == "CONSENS":
                current_consensus_id += 1
                if specific_consensuses_ids is None or current_consensus_id in specific_consensuses_ids:
                    detailed_consens_info_line = next(po_lines_iterator)
                    detailed_consens_info = self._extract_line_value(detailed_consens_info_line).split(' ')
                    consens_nodes_count = int(detailed_consens_info[0])
                    cons_id = int(path_name[7:])
                    if cons_id in consensuses_in_po_lines:
                        consensuses_in_po_lines[cons_id].fullname = path_name
                        consensuses_in_po_lines[cons_id].po_consensus_id = f"S{str(i)}"
                        consensuses_in_po_lines[cons_id].path = [None] * consens_nodes_count
                    else:
                        consensuses_in_po_lines[cons_id] = ConsInfo(fullname=f"CONSENS{cons_id}",
                                                                    assigned_sequences_ids=[],
                                                                    po_consensus_id=f"S{str(i)}",
                                                                    path=[None] * consens_nodes_count)
            else:
                detailed_sequence_info_line = next(po_lines_iterator)
                detailed_sequence_info = self._extract_line_value(detailed_sequence_info_line).split(' ')
                assigned_consensus_id = int(detailed_sequence_info[3])
                sequence_id = self.seq_new_to_old[i]
                if assigned_consensus_id != -1:
                    if assigned_consensus_id in consensuses_in_po_lines:
                        consensuses_in_po_lines[assigned_consensus_id].assigned_sequences_ids.append(sequence_id)
                    else:
                        consensuses_in_po_lines[assigned_consensus_id] = ConsInfo(fullname=f"CONSENS{assigned_consensus_id}",
                                                                                  assigned_sequences_ids=[sequence_id])

        if not consensuses_in_po_lines:
            raise NoConsensusError("No consensus found in this poagraph.")

        consensuses_paths_node_counter = {c: 0
                                          for c in consensuses_in_po_lines.keys()}

        first_node_line_idx = 5 + paths_count * 2
        new_node_id = -1

        for line_idx in range(first_node_line_idx, len(po_lines)):
            new_node_id += 1
            for c_id in range(len(consensuses_in_po_lines)):
                if consensuses_in_po_lines[c_id].po_consensus_id in po_lines[line_idx]:
                    consensuses_in_po_lines[c_id].path[consensuses_paths_node_counter[c_id]] = self.new_to_old[new_node_id]
                    consensuses_paths_node_counter[c_id] += 1

        return consensuses_in_po_lines



import unittest
from pathlib import Path

from ddt import unpack, data, ddt

from pangtreebuild.pangenome import graph
from pangtreebuild.serialization import po
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import pathtools


def nid(x): return graph.NodeID(x)


def bid(x): return graph.Base(x)


@ddt
class PoagraphToPOTests(unittest.TestCase):

    def setUp(self):
        self.po_files_dir = Path(__file__).parent.joinpath('po_files')

    def test_1_typical_poagraph(self):
        expected_po_content_path = self.po_files_dir.joinpath("test_1.po")

        poagraph_nodes = [graph.Node(node_id=nid(0), base=bid('A'), aligned_to=nid(1)),
                          graph.Node(node_id=nid(1), base=bid('G'), aligned_to=nid(0)),
                          graph.Node(node_id=nid(2), base=bid('C'), aligned_to=nid(3)),
                          graph.Node(node_id=nid(3), base=bid('G'), aligned_to=nid(2)),
                          graph.Node(node_id=nid(4), base=bid('A'), aligned_to=nid(5)),
                          graph.Node(node_id=nid(5), base=bid('T'), aligned_to=nid(4)),
                          graph.Node(node_id=nid(6), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(7), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(8), base=bid('A'), aligned_to=nid(9)),
                          graph.Node(node_id=nid(9), base=bid('C'), aligned_to=nid(10)),
                          graph.Node(node_id=nid(10), base=bid('G'), aligned_to=nid(11)),
                          graph.Node(node_id=nid(11), base=bid('T'), aligned_to=nid(8)),
                          graph.Node(node_id=nid(12), base=bid('A'), aligned_to=nid(13)),
                          graph.Node(node_id=nid(13), base=bid('C'), aligned_to=nid(12)),
                          graph.Node(node_id=nid(14), base=bid('T'), aligned_to=None),
                          graph.Node(node_id=nid(15), base=bid('A'), aligned_to=nid(16)),
                          graph.Node(node_id=nid(16), base=bid('C'), aligned_to=nid(17)),
                          graph.Node(node_id=nid(17), base=bid('G'), aligned_to=nid(15))
                          ]

        poagraph_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 2, 4, 6, 7, 8, 12, 14, 16])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 2, 5, 6, 7, 9])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [3, 4, 6, 7, 10, 12, 14, 17])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [11, 13, 14, 15])])],
                               graph.SequenceMetadata({'group': '1'})),
        }

        poagraph = graph.Poagraph(poagraph_nodes, poagraph_sequences)

        actual_po_content = po.poagraph_to_PangenomePO(poagraph)
        expected_po_content = pathtools.get_file_content(expected_po_content_path)
        self.assertEqual(expected_po_content, actual_po_content)

    def test_2_consensuses_and_empty_sequences(self):
        expected_po_content_path = self.po_files_dir.joinpath("test_2.po")

        poagraph_nodes = [graph.Node(node_id=nid(0), base=bid('C'), aligned_to=nid(1)),
                          graph.Node(node_id=nid(1), base=bid('T'), aligned_to=nid(0)),
                          graph.Node(node_id=nid(2), base=bid('A'), aligned_to=nid(3)),
                          graph.Node(node_id=nid(3), base=bid('G'), aligned_to=nid(2)),
                          graph.Node(node_id=nid(4), base=bid('C'), aligned_to=None),
                          graph.Node(node_id=nid(5), base=bid('T'), aligned_to=None),
                          graph.Node(node_id=nid(6), base=bid('A'), aligned_to=nid(7)),
                          graph.Node(node_id=nid(7), base=bid('T'), aligned_to=nid(6)),
                          graph.Node(node_id=nid(8), base=bid('G'), aligned_to=None)
                          ]

        poagraph_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 3, 4, 5, 6, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 2, 4, 5, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('CONSENS0'):
                graph.Sequence(msa.SequenceID('CONSENS0'),
                               [graph.SeqPath([*map(nid, [0, 3, 4, 5, 7, 8])])],
                               None),
            msa.SequenceID('CONSENS1'):
                graph.Sequence(msa.SequenceID('CONSENS1'),
                               [graph.SeqPath([*map(nid, [1, 2, 4, 5, 6, 8])])],
                               None),
        }

        poagraph = graph.Poagraph(poagraph_nodes, poagraph_sequences)

        actual_po_content = po.poagraph_to_PangenomePO(poagraph)
        expected_po_content = pathtools.get_file_content(expected_po_content_path)
        self.assertEqual(expected_po_content, actual_po_content)

    @data(([], ""),
          ([nid(0), nid(1)], "L0L1"))
    @unpack
    def test_3_get_in_nodes_info(self, in_nodes, expected_in_nodes):
        actual_in_nodes = po._get_in_nodes_info(in_nodes)
        self.assertEqual(expected_in_nodes, actual_in_nodes)

    @data(([], ""),
          ([0, 1], "S0S1"),
          ([1, 3], "S1S3"))
    @unpack
    def test_get_sources_info(self, sources_ids, expected_info):
        actual_sources_ids = po._get_sources_info(sources_ids)
        self.assertEqual(expected_info, actual_sources_ids)

    @data((None, ""),
          (0, "A0"))
    @unpack
    def test_get_aligned_to_info(self, aligned_to, expected_aligned_to):
        actual_aligned_to = po._get_aligned_to_info(aligned_to)
        self.assertEqual(expected_aligned_to, actual_aligned_to)


if __name__ == '__main__':
    unittest.main()



import setuptools


setuptools.setup(
    name='pangtreebuild',
    version='1.0.0',
    author="Paulina Dziadkiewicz",
    author_email="pedziadkiewicz@gmail.com",
    description="Multiple sequence alignment analysis with Affinity Tree generation",
    url="https://github.com/meoke/pangtree",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    install_requires=['biopython', 'numpy', 'jsonpickle', 'ddt', 'networkx', 'six', 'newick'],
    license='MIT Licence',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
)



import argparse
import inspect
import os
from io import StringIO
from pathlib import Path
from typing import TypeVar, Callable, Optional, Union

from pangtreebuild.serialization.json import TaskParameters
from pangtreebuild.affinity_tree import parameters as at_params
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.tools import pathtools


class InvalidPath(Exception):
    pass


def _get_file_extension(arg: str) -> str:
    """Returns file extension if present in arg else throws InvalidPath."""

    file_path_suffix = Path(arg).suffix
    try:
        return file_path_suffix.split('.')[1]
    except Exception:
        raise InvalidPath(f"Cannot find file extension in {arg}.")


def _data_type(data_type: str) -> graph.DataType:
    """Converts command line argument to DataType"""

    if data_type == "p":
        data_type_full_name = "Proteins"
    elif data_type == "n":
        data_type_full_name = "Nucleotides"
    else:
        raise argparse.ArgumentError("""Unknown data type. \'p\' for proteins
                                        or \'n\' for nucleotides available.""")
    try:
        dt = graph.DataType[data_type_full_name]
        return dt
    except KeyError:
        raise argparse.ArgumentError("Data type parsing error.")


def _path_if_valid(path: str) -> Path:
    """Check if path exists."""

    file_path = Path(path)
    if not pathtools.file_exists(file_path):
        raise InvalidPath(f"{file_path}")
    return file_path


def _cli_dir_arg(path: str) -> Path:
    """Check if dir exists and creates it if not."""

    dir_path = Path(path)
    if not pathtools.dir_exists(dir_path):
        pathtools.create_dir(dir_path)
    return dir_path


T = TypeVar('T')


def _cli_file_arg(arg: str,
                  constructor: Callable[[StringIO, Optional[Path]], T]) -> \
        T:
    try:
        filepath = _path_if_valid(arg)
    except InvalidPath:
        raise argparse.ArgumentTypeError(f"""File {arg} does not exist
                                             or is not a file.""")
    with open(arg) as infile:
        file_content = StringIO(infile.read())
        try:
            return constructor(file_content, filepath)
        except Exception as p:
            raise argparse.ArgumentError("Incorrect file content") from p


def _mulitalignment_file(x: str) -> Union[msa.Maf, msa.Po]:
    file_type = _get_file_extension(x)
    if file_type == 'maf':
        return _cli_file_arg(x, msa.Maf)
    elif file_type == 'po':
        return _cli_file_arg(x, msa.Po)
    else:
        raise InvalidPath("""Only msa files with
                            .maf or .po can be processed.""")


def _metadata_file(x: str) -> msa.MetadataCSV:
    return _cli_file_arg(x, msa.MetadataCSV)


def _blosum_file(x: str) -> at_params.Blosum:
    return _cli_file_arg(x, at_params.Blosum)


def _cli_arg(constructor: Callable[[str], T]) -> Callable[[str], T]:
    def _c(x):
        try:
            return constructor(x)
        except Exception as p:
            raise argparse.ArgumentError(f"Incorrect argument {x}") from p
    return _c


def get_parser() -> argparse.ArgumentParser:
    """Create ArgumentParser for pang module."""

    p = argparse.ArgumentParser(prog='pangtreebuild',
                                description="""This software builds poagraph
                                                and generates affinitytree.""",
                                epilog="""For more information check
                                          github.com/meoke/pangtree""")
    p.add_argument('--output_dir',
                   type=_cli_dir_arg,
                   default=get_default_output_dir(),
                   help='Output directory path.')
    p.add_argument('--multialignment',
                   metavar='MULTIALIGNMENT_PATH',
                   type=_mulitalignment_file,
                   required=True,
                   help='Path to the multialignment file.')
    p.add_argument('--datatype',
                   type=_data_type,
                   default=graph.DataType.Nucleotides,
                   help='\'n\' for nucleotides, \'p\' for proteins. ' +
                        inspect.getdoc(graph.DataType))
    p.add_argument('--metadata',
                   metavar='METADATA_PATH',
                   type=_metadata_file,
                   help='Path to the csv file with metadata. ' +
                        inspect.getdoc(msa.MetadataCSV))
    p.add_argument('--raw_maf',
                   action='store_true',
                   default=False,
                   help="""Poagraph building from maf file parameter. Set if
                           the maf content must not be transformed to DAG
                           before building graph. Poagraph that was build
                           in this way provides affinitytree tree but the
                           affinitytree do not reflect the real life
                           sequences.""")
    p.add_argument('--fasta_provider',
                   metavar="FASTA_PROVIDER",
                   choices=['ncbi', 'file'],
                   help="""'Maf file may not include full _sequences.
                            In such case an additional data source is needed.
                            Use \'ncbi\' for NCBI (activates CACHE option)
                            or \'file\' for file (then provide also
                            FASTA_PATH). MISSING_SYMBOL is used if this
                            argument is omitted.""")
    p.add_argument('--missing_symbol',
                   metavar='MISSING_SYMBOL',
                   type=_cli_arg(missings.MissingBase),
                   help=inspect.getdoc(missings.MissingBase))
    p.add_argument('--cache',
                   action='store_true',
                   help="""Set if fastas downloaded from NCBI should be cached
                           locally in .fastacache folder. Used if Fasta
                           Provider is NCBI. Sequences downloaded from NCBI
                           are stored and reused by this program.""")
    p.add_argument('--fasta_path',
                   metavar="FASTA_PATH",
                   type=_path_if_valid,
                   help="""ZIP archive with fasta files or fasta file used
                        as FASTA_PROVIDER.""")
    p.add_argument('--affinity',
                   choices=['poa', 'tree'],
                   help="""Generate affinity tree. Use \'poa\' for direct
                           result of poa software, \'tree\' for Affinity
                           Tree algorithm.""")
    p.add_argument('--blosum',
                   type=_blosum_file,
                   metavar='BLOSUM_PATH',
                   help='Path to the blosum file. ' +
                        inspect.getdoc(at_params.Blosum))
    p.add_argument('--hbmin',
                   type=_cli_arg(at_params.Hbmin),
                   default=at_params.Hbmin(),
                   help='Simple POA algorithm parameter. '
                        'Hbmin value. ' + inspect.getdoc(at_params.Hbmin))
    p.add_argument('--stop',
                   type=_cli_arg(at_params.Stop),
                   default=at_params.Stop(),
                   help='Tree POA algorithm parameter.' +
                        inspect.getdoc(at_params.Stop))
    p.add_argument('--p',
                   type=_cli_arg(at_params.P),
                   default=at_params.P(),
                   help='Tree consensus algorithm parameter.' +
                        inspect.getdoc(at_params.P))
    p.add_argument('--output_fasta',
                   action='store_true',
                   help="""Set if fasta files for _sequences and
                            affinitytree must be produced.""")
    p.add_argument('--output_po',
                   action='store_true',
                   default=False,
                   help='Set if po file for poagraph must be produced.')
    p.add_argument('-v', '--verbose',
                   action='store_true',
                   default=False,
                   help='Set if detailed log files must be produced.')
    p.add_argument('-q', '--quiet',
                   action='store_true',
                   default=False,
                   help='Set to turn off console logging.')
    return p


def resolve_fasta_provider(args: argparse.Namespace) -> \
        missings.FastaProvider:
    """Returns fasta provider based on parsed arguments."""

    if args.fasta_provider is None:
        if args.missing_symbol is None:
            return missings.ConstBaseProvider(missings.MissingBase())
        else:
            return missings.ConstBaseProvider(args.missing_symbol)
    elif args.fasta_provider == 'ncbi':
        use_cache = args.cache if args.cache else False
        return missings.FromNCBI(use_cache)
    elif args.fasta_provider == 'file':
        if args.fasta_path is None:
            raise Exception("""Fasta file source must be specified.
                               It must be provided when fasta source
                               is \'local\'.""")
        return missings.FromFile(args.fasta_path)
    else:
        raise Exception("""Not known fasta provider.
                           Should be \'ncbi\' or \'file\' or None.
                           Cannot build graph.""")


def get_default_output_dir():
    """Creates timestamped child dir under current working directory."""

    current_dir = pathtools.get_cwd()
    output_dir = pathtools.get_child_path(current_dir, "output")
    pathtools.create_dir(output_dir)
    current_time = pathtools.get_current_time()
    output_dir_name = "_".join(["output", current_time])
    output_dir_path = pathtools.get_child_path(output_dir,
                                               output_dir_name)
    pathtools.create_dir(output_dir_path)
    return output_dir_path


def get_default_blosum():
    """Returns default blosum file: Blosum80.mat"""

    pangtreebuild_dir = Path(__file__).parent.parent
    default_blosum_path = pathtools.get_child_path(pangtreebuild_dir, "affinity_tree/bin/blosum80.mat")
    blosum_content = pathtools.get_file_content_stringio(default_blosum_path)
    return at_params.Blosum(blosum_content, default_blosum_path)


def get_task_parameters(args: argparse.Namespace, running_time) -> \
        TaskParameters:
    """Returns TaskParameters object based on parsed arguments."""

    return TaskParameters(running_time=running_time,
                          multialignment_file_path=args.multialignment.filename,
                          multialignment_format=str(type(args.multialignment).__name__),
                          datatype=args.datatype.name,
                          metadata_file_path=args.metadata.filename if args.metadata else None,
                          blosum_file_path=args.blosum.filepath if args.blosum else None,
                          output_path=args.output_dir,
                          output_po=bool(args.output_po),
                          output_fasta=bool(args.output_fasta),
                          output_with_nodes=True,
                          verbose=bool(args.verbose),
                          raw_maf=bool(args.raw_maf),
                          fasta_provider=args.fasta_provider if args.fasta_provider else 'ConstSymbol',
                          cache=bool(args.cache),
                          missing_base_symbol=args.missing_symbol.value if args.missing_symbol else missings.MissingBase().value,
                          fasta_source_file=str(args.fasta_path),
                          consensus_type=args.affinity,
                          hbmin=args.hbmin.value if args.hbmin else None,
                          stop=args.stop.value if args.stop else None,
                          p=args.p.value if args.p else None)



# -*- coding: utf-8 -*-
import six
from .graph.EdgeInfo import EdgeInfo
from collections import defaultdict, namedtuple

Edge = namedtuple('Edge', ['left', 'right', 'type'])

def edge_type(strand1, strand2):
    return (strand1, -strand2)
    
def connect_blocks(seq1, seq2, d):
    edgeType = edge_type(seq1.strand, seq2.strand)
    if seq1.block == seq2.block: 
        edgeType = edge_type(min(seq1.strand, seq2.strand), max(seq1.strand, seq2.strand))    
    elif seq1.block > seq2.block:
        edgeType = edgeType[::-1]
        seq1, seq2 = seq2, seq1   
    e = Edge(seq1.block, seq2.block, edgeType)
    d[e][0].append((EdgeInfo(seq1.id, seq1.start), EdgeInfo(seq2.id, seq2.start)))
    d[e][1] += 1

def weight(seq):
    d = defaultdict(lambda: [[], 0])
    if six.PY2:
        for i in xrange(len(seq) - 1):
            if seq[i].id == seq[i+1].id:
                connect_blocks(seq[i], seq[i+1], d)
    else:
        for i in range(len(seq) - 1):
            if seq[i].id == seq[i+1].id:
                connect_blocks(seq[i], seq[i+1], d)
    return d
